#include <stdio.h>
#include <stdlib.h>

//************************************************************************
// This program reads two Spin binary files and links them together.
// The first file is a parent binary object file.  It must contain two
// objects, which are the top object and a stub of the child object.
//
// The second file is the child binary object file.  It must contain a
// single object whose PUB and PRI methods contain the same order and
// number of calling parameters as the stub object used in the parent
// file.
//
// This program concatenates the first objects from the parent and child
// binary files, and writes out a new binary file.  The 16-byte program
// header is adjusted to match up with the new program file containing
// the two objects.
//
// Dave Hein, January 31, 2011
//************************************************************************

// Define a 32K buffer
static unsigned short buffer[16384];

void usage(void)
{
    printf("usage: linkit parent.binary child.binary out.binary\n");
    exit(1);
}

FILE *OpenFile(char *fname, char *mode)
{
    FILE *stream = fopen(fname, mode);

    if (!stream)
    {
	printf("Could not open %s\n", fname);
	exit(1);
    }

    return stream;
}

void main(int argc, char **argv)
{
    int i;
    FILE *infile;
    FILE *outfile;
    int stackoff;
    int varsize1, varsize2;
    int objsize1, objsize2;
    int checksum = 0;
    unsigned char *byteptr = (unsigned char *)buffer;

    if (argc != 4) usage();

    // Read the parent file and print the object information
    infile = OpenFile(argv[1], "rb");
    fread(buffer, 1, 32768, infile);
    fclose(infile);
    printf("Parent object %s\n", argv[1]);
    printf("pbase = %4.4x, vbase = %4.4x, dbase = %4.4x, pcurr = %4.4x, dcurr = %4.4x\n",
        buffer[3], buffer[4], buffer[5], buffer[6], buffer[7]);
    printf("Object size = %4.4x, ", buffer[8]);
    printf("VAR size = %4.4x\n", buffer[5] - buffer[4] - 8);

    // Extract the object size, VAR size and stack offset
    objsize1 = buffer[8];
    varsize1 = buffer[5] - buffer[4] - 8;
    stackoff = buffer[7] - buffer[5];

    printf("\n");

    // Read the child file and print the object information
    infile = OpenFile(argv[2], "rb");
    fread(buffer, 1, 32768, infile);
    fclose(infile);
    printf("Child object %s\n", argv[2]);
    printf("pbase = %4.4x, vbase = %4.4x, dbase = %4.4x, pcurr = %4.4x, dcurr = %4.4x\n",
        buffer[3], buffer[4], buffer[5], buffer[6], buffer[7]);
    printf("Object size = %4.4x, ", buffer[8]);
    printf("VAR size = %4.4x\n", buffer[5] - buffer[4] - 8);

    // Extract the object size and VAR size
    objsize2 = buffer[8];
    varsize2 = buffer[5] - buffer[4] - 8;

    // Re-read the header plus the first object in the parent file
    infile = OpenFile(argv[1], "rb");
    fread(buffer, 1, objsize1 + 16, infile);
    fclose(infile);

    // Skip the child's header and read in the first object
    infile = OpenFile(argv[2], "rb");
    fread(((char *)buffer) + objsize1 + 16, 1, 16, infile);
    fread(((char *)buffer) + objsize1 + 16, 1, objsize2, infile);
    fclose(infile);

    // Set VBASE in the header to the sum of the object size
    // plus the header size
    buffer[4] = objsize1 + objsize2 + 16;

    // Set DBASE in the header to the VBASE plus the sum of
    // the VAR sizes plus 8 for the top return frame
    buffer[5] = buffer[4] + varsize1 + varsize2 + 8;

    // Set DCURR in the header to DBASE plus the stack offset.
    // The stack offset is normally 4, but it could be larger if
    // the first method contains calling parameters.
    buffer[7] = buffer[5] + stackoff;

    // Compute the new checksum and write it at header byte 5
    for (i = 0; i < buffer[5]; i++) checksum += byteptr[i];
    byteptr[5] += 0x14 - checksum;

    printf("\n");

    // Print the new program values
    printf("Linked program %s\n", argv[3]);
    printf("pbase = %4.4x, vbase = %4.4x, dbase = %4.4x, pcurr = %4.4x, dcurr = %4.4x\n",
        buffer[3], buffer[4], buffer[5], buffer[6], buffer[7]);
    printf("Object size = %4.4x, ", buffer[8]);
    printf("VAR size = %4.4x\n", buffer[5] - buffer[4] - 8);

    // Write the output file
    outfile = OpenFile(argv[3], "wb");
    fwrite(buffer, 1, buffer[4], outfile);
    fclose(outfile);
}
